#include "LibavStreamer.h"

#include <windows.h>

extern "C"
{
#include <libavcodec/avcodec.h>
#include <libavformat/avformat.h>
#include <libavfilter/buffersink.h>
#include <libavfilter/buffersrc.h>
#include <libavutil/channel_layout.h>
#include <libavutil/opt.h>
#include <libavdevice/avdevice.h>
#include <libavutil/audio_fifo.h>
}

LibavStreamer::LibavStreamer()
{
	::CoInitializeEx(nullptr, COINIT_MULTITHREADED);

	avformat_network_init();
	avdevice_register_all();
}

LibavStreamer::~LibavStreamer()
{
	avformat_network_deinit();

	::CoUninitialize();
}

bool LibavStreamer::OpenMic(std::wstring micName)
{
	std::wstring dshowInput = L"audio=" + micName;
	return OpenDShow(dshowInput);
}

bool LibavStreamer::OpenCam(std::wstring camName)
{
	std::wstring dshowInput = L"video=" + camName;
	return OpenDShow(dshowInput);
}

bool LibavStreamer::OpenCamMic(std::wstring camName, std::wstring micName)
{
	std::wstring dshowInput = L"video=" + camName + L":audio=" + micName;
	return OpenDShow(dshowInput);
}

std::string LibavStreamer::ConvertFromWString(std::wstring wstr)
{
	int length = wstr.length() + 1;
	int sizeNeeded = WideCharToMultiByte(CP_UTF8, 0, wstr.c_str(), length, nullptr, 0, nullptr, nullptr);

	char* strTo = (char*)malloc(sizeNeeded);
	if (strTo == nullptr)
	{
		return "";
	}

	ZeroMemory(strTo, sizeNeeded);

	WideCharToMultiByte(CP_UTF8, 0, wstr.c_str(), wstr.length(), strTo, sizeNeeded, nullptr, nullptr);

	std::string outputString(strTo);

	free(strTo);

	return outputString;
}

bool LibavStreamer::OpenDShow(std::wstring dshowInputString)
{
	AVInputFormat* pInputFormat = av_find_input_format("dshow");
	if (pInputFormat == nullptr)
	{
		return false;
	}

	AVDictionary* pDShowOptions = nullptr;

	av_dict_set(&pDShowOptions, "rtbufsize", "702000k", 0);
	av_dict_set(&pDShowOptions, "video_size", "640x480", 0);
	av_dict_set(&pDShowOptions, "framerate", "30", 0);
	av_dict_set(&pDShowOptions, "channels", "1", 0);
	av_dict_set(&pDShowOptions, "sample_rate", "11025", 0);//22050
	av_dict_set(&pDShowOptions, "audio_buffer_size", "80", 0);

	int result = avformat_open_input(&m_pInputFormatContext, ConvertFromWString(dshowInputString).c_str(), pInputFormat, &pDShowOptions);

	if (result < 0)
	{
		av_dict_free(&pDShowOptions);

		av_log(NULL, AV_LOG_ERROR, "Cannot open Direct Show input.\n");

		return false;
	}

	av_dict_free(&pDShowOptions);

	result = avformat_find_stream_info(m_pInputFormatContext, nullptr);
	if (result < 0)
	{
		av_log(nullptr, AV_LOG_ERROR, "Cannot find input stream information\n");
		return false;
	}

	for (int streamIndex = 0; streamIndex < m_pInputFormatContext->nb_streams; ++streamIndex)
	{
		AVStream* pAvStream = m_pInputFormatContext->streams[streamIndex];

		const AVCodec* pDecoder = avcodec_find_decoder(pAvStream->codecpar->codec_id);
		if (pDecoder == nullptr)
		{
			av_log(nullptr, AV_LOG_ERROR, "Failed to find decoder for stream #%u\n", streamIndex);
			return false;
		}

		AVCodecContext* pCodecContext = avcodec_alloc_context3(pDecoder);
		if (pCodecContext == nullptr)
		{
			av_log(NULL, AV_LOG_ERROR, "Failed to allocate the decoder context for stream #%u\n", streamIndex);
			return false;
		}

		result = avcodec_parameters_to_context(pCodecContext, pAvStream->codecpar);
		if (result < 0)
		{
			av_log(nullptr, AV_LOG_ERROR, "Failed to copy decoder parameters to input decoder context for stream #%u\n", streamIndex);
			return false;
		}

		if (pCodecContext->codec_type == AVMEDIA_TYPE_VIDEO || pCodecContext->codec_type == AVMEDIA_TYPE_AUDIO)
		{
			if (pCodecContext->codec_type == AVMEDIA_TYPE_VIDEO)
			{
				const AVRational dst_fps = { 30, 1 };
				pCodecContext->framerate = dst_fps;
			}

			result = avcodec_open2(pCodecContext, pDecoder, nullptr);
			if (result < 0)
			{
				av_log(nullptr, AV_LOG_ERROR, "Failed to open decoder for stream #%u\n", streamIndex);
				return false;
			}
		}

		StreamContext streamContext;
		streamContext.pDecodingContext = pCodecContext;

		streamContext.pPacketsQueue = new moodycamel::ReaderWriterQueue<AVPacket*>();

		m_streamContexts.push_back(streamContext);
	}

	return true;
}

namespace
{
	void PrintAvailableEncoders()
	{
		void* iterate_data = NULL;
		const AVCodec* current_codec = av_codec_iterate(&iterate_data);
		while (current_codec != NULL)
		{
			std::string codecName = std::string("Name: ") + std::string(current_codec->name) + std::string(" Long Name: ") + std::string(current_codec->long_name) + std::string("\n");
			av_log(nullptr, AV_LOG_INFO, codecName.c_str());
			current_codec = av_codec_iterate(&iterate_data);
		}
	}
}

bool LibavStreamer::OpenOutputUrl(std::wstring url)
{
	if (m_pInputFormatContext == nullptr)
	{
		av_log(nullptr, AV_LOG_ERROR, "Open Direct Show input first!\n");
		return false;
	}

	std::string urlString = ConvertFromWString(url);
	AVOutputFormat* pOutputFormat = av_guess_format("flv", urlString.c_str(), nullptr);
	if (pOutputFormat == nullptr)
	{
		return false;
	}

	int result = avformat_alloc_output_context2(&m_pOutputFormatContext, pOutputFormat, nullptr, urlString.c_str());
	if (result < 0)
	{
		av_log(nullptr, AV_LOG_ERROR, "Could not create output context\n");
		return false;
	}

	for (int streamIndex = 0; streamIndex < m_pInputFormatContext->nb_streams; ++streamIndex)
	{
		if (!(m_pInputFormatContext->streams[streamIndex]->codecpar->codec_type == AVMEDIA_TYPE_AUDIO
			|| m_pInputFormatContext->streams[streamIndex]->codecpar->codec_type == AVMEDIA_TYPE_VIDEO))
		{
			continue;
		}

		AVStream* pOutputStream = avformat_new_stream(m_pOutputFormatContext, nullptr);
		if (pOutputStream == nullptr)
		{
			av_log(nullptr, AV_LOG_ERROR, "Failed allocating output stream\n");
			return false;
		}

		AVStream* pInputStream = m_pInputFormatContext->streams[streamIndex];

		AVCodecContext* pDecodingContext = m_streamContexts[streamIndex].pDecodingContext;

		if (pDecodingContext->codec_type == AVMEDIA_TYPE_VIDEO || pDecodingContext->codec_type == AVMEDIA_TYPE_AUDIO)
		{
			const AVCodec* pEncoder = nullptr;

			if (pDecodingContext->codec_type == AVMEDIA_TYPE_VIDEO)
			{
				//pEncoder = avcodec_find_encoder(AV_CODEC_ID_H264);
				pEncoder = avcodec_find_encoder_by_name("h264_mf");
			}

			if (pDecodingContext->codec_type == AVMEDIA_TYPE_AUDIO)
			{
				//pEncoder = avcodec_find_encoder_by_name("aac_mf");
				pEncoder = avcodec_find_encoder(AV_CODEC_ID_AAC);
			}

			if (pEncoder == nullptr)
			{
				av_log(nullptr, AV_LOG_FATAL, "Necessary encoder not found\n");
				return false;
			}

			AVCodecContext* pEncodingContext = avcodec_alloc_context3(pEncoder);
			if (pEncodingContext == nullptr)
			{
				av_log(nullptr, AV_LOG_FATAL, "Failed to allocate the encoder context\n");
				return false;
			}

			if (pDecodingContext->codec_type == AVMEDIA_TYPE_VIDEO)
			{
				pEncodingContext->codec_tag = 0;
				pEncodingContext->width = pDecodingContext->width / 2;
				pEncodingContext->height = pDecodingContext->height / 2;
				pEncodingContext->sample_aspect_ratio = pDecodingContext->sample_aspect_ratio;
				pEncodingContext->pix_fmt = pEncoder->pix_fmts[0];//AV_PIX_FMT_YUV420P;//AV_PIX_FMT_YUV422P;
				pEncodingContext->bit_rate = 128000;//256000;///512000;

				const AVRational dst_fps = { 30, 1 };

				pEncodingContext->framerate = dst_fps;
				pEncodingContext->time_base = av_inv_q(dst_fps);
				pEncodingContext->gop_size = 15;
				//pEncodingContext->keyint_min = 5;
			}

			if (pDecodingContext->codec_type == AVMEDIA_TYPE_AUDIO)
			{
				pEncodingContext->sample_rate = pDecodingContext->sample_rate;
				pEncodingContext->channel_layout = av_get_default_channel_layout(1);
				pEncodingContext->channels = 1;
				pEncodingContext->sample_fmt = pEncoder->sample_fmts[0];

				AVRational rational;
				rational.num = 1;
				rational.den = pEncodingContext->sample_rate;

				pEncodingContext->time_base = rational;

				pEncodingContext->bit_rate = 64000;

				//pEncodingContext->strict_std_compliance = FF_COMPLIANCE_EXPERIMENTAL;
			}

			m_streamContexts[streamIndex].pEncodingContext = pEncodingContext;

			if (m_pOutputFormatContext->oformat->flags & AVFMT_GLOBALHEADER)
			{
				pEncodingContext->flags |= AV_CODEC_FLAG_GLOBAL_HEADER;
			}

			if (pDecodingContext->codec_type == AVMEDIA_TYPE_VIDEO)
			{
				//AVDictionary* pCodecOptions = nullptr;

				//av_dict_set(&pCodecOptions, "preset", "ultrafast", 0);
				//av_dict_set(&pCodecOptions, "tune", "zerolatency", 0);

				//av_dict_set(&pCodecOptions, "allow_skip_frames", "1", 0);
				//av_dict_set(&pCodecOptions, "loopfilter", "0", 0);
				//av_dict_set(&pCodecOptions, "maxrate", "512000", 0);

				AVDictionary* pCodecOptions = nullptr;

				av_dict_set(&pCodecOptions, "scenario", "video_conference", 0);
				av_dict_set(&pCodecOptions, "rate_control", "cbr", 0);
				//av_dict_set(&pCodecOptions, "hw_encoding", "1", 0);

				result = avcodec_open2(pEncodingContext, pEncoder, &pCodecOptions);

				//av_dict_free(&pCodecOptions);
			}

			if (pDecodingContext->codec_type == AVMEDIA_TYPE_AUDIO)
			{
				AVDictionary* pCodecOptions = nullptr;

				av_dict_set(&pCodecOptions, "aac_coder", "fast", 0);
				//av_dict_set(&pCodecOptions, "bufsize", "1024k", 0);

				result = avcodec_open2(pEncodingContext, pEncoder, &pCodecOptions);

				av_dict_free(&pCodecOptions);
			}

			if (result < 0)
			{
				av_log(nullptr, AV_LOG_ERROR, "Cannot open encoder for stream #%u\n", streamIndex);
				return false;
			}

			result = avcodec_parameters_from_context(pOutputStream->codecpar, pEncodingContext);
			if (result < 0)
			{
				av_log(nullptr, AV_LOG_ERROR, "Failed to copy encoder parameters to output stream #%u\n", streamIndex);
				return result;
			}

			pOutputStream->time_base = pEncodingContext->time_base;
		}
	}

	m_pOutputFormatContext->oformat->flags |= AVFMT_ALLOW_FLUSH;
	m_pOutputFormatContext->flags |= AVFMT_FLAG_FLUSH_PACKETS;
	m_pOutputFormatContext->flags |= AVFMT_FLAG_NOBUFFER;
	//m_pOutputFormatContext->flush_packets = 0;
	//m_pOutputFormatContext->max_delay = 0;
	//m_pOutputFormatContext->start_time_realtime = AV_NOPTS_VALUE;

	if (!(m_pOutputFormatContext->oformat->flags & AVFMT_NOFILE))
	{
		result = avio_open(&m_pOutputFormatContext->pb, urlString.c_str(), AVIO_FLAG_WRITE);
		if (result < 0)
		{
			av_log(NULL, AV_LOG_ERROR, "Could not open output URL '%s'", urlString);
			return false;
		}
	}

	AVDictionary* pHeaderOptions = nullptr;
	av_dict_set(&pHeaderOptions, "flush_packets", "1", 0);
	av_dict_set(&pHeaderOptions, "tcp_nodelay", "1", 0);
	av_dict_set(&pHeaderOptions, "max_delay", "1", 0);

	result = avformat_write_header(m_pOutputFormatContext, &pHeaderOptions);

	av_dict_free(&pHeaderOptions);

	if (result < 0)
	{
		av_log(nullptr, AV_LOG_ERROR, "Error occurred when opening output URL\n");
		return false;
	}

	return true;
}

void LibavStreamer::Cleanup()
{
	if (pReceivingPacket != nullptr)
	{
		av_packet_free(&pReceivingPacket);
		pReceivingPacket = nullptr;
	}

	for (auto&& streamContext : m_streamContexts)
	{
		if (streamContext.pDecodingContext != nullptr)
		{
			avcodec_free_context(&streamContext.pDecodingContext);
		}

		if (streamContext.pEncodingContext != nullptr)
		{
			avcodec_free_context(&streamContext.pEncodingContext);
		}

		if (streamContext.pPacketsQueue != nullptr)
		{
			AVPacket* pPacket = nullptr;
			while (streamContext.pPacketsQueue->try_dequeue<AVPacket*>(pPacket))
			{
				av_packet_free(&pPacket);
			}
			delete streamContext.pPacketsQueue;
			streamContext.pPacketsQueue = nullptr;
		}
	}

	m_streamContexts.clear();

	for (auto&& filteringContext : m_filteringContexts)
	{
		if (filteringContext.pFilterGraph != nullptr)
		{
			avfilter_graph_free(&filteringContext.pFilterGraph);
		}
		if (filteringContext.pEncodedPacket != nullptr)
		{
			av_packet_free(&filteringContext.pEncodedPacket);
		}
		if (filteringContext.pFilteredFrame != nullptr)
		{
			av_frame_free(&filteringContext.pFilteredFrame);
		}
	}

	m_filteringContexts.clear();

	if (m_pInputFormatContext != nullptr)
	{
		avformat_close_input(&m_pInputFormatContext);
		m_pInputFormatContext = nullptr;
	}

	if (m_pOutputFormatContext != nullptr)
	{
		if (!(m_pOutputFormatContext->oformat->flags & AVFMT_NOFILE))
		{
			avio_closep(&m_pOutputFormatContext->pb);
		}
		avformat_free_context(m_pOutputFormatContext);
		m_pOutputFormatContext = nullptr;
	}

	if (pFifo != nullptr)
	{
		av_audio_fifo_free(pFifo);
		pFifo = nullptr;
	}
}

bool LibavStreamer::InitFilter(AVCodecContext* pDecodingContext, AVCodecContext* pEncodingContext, const char* pFilterSpec)
{
	AVFilterGraph* pFilterGraph = avfilter_graph_alloc();
	if (pFilterGraph == nullptr)
	{
		return false;
	}

	AVFilterContext* pBufferSourceContext = nullptr;
	AVFilterContext* pBufferSinkContext = nullptr;

	if (pDecodingContext->codec_type == AVMEDIA_TYPE_VIDEO)
	{
		char args[512];
		snprintf(args, sizeof(args),
			"video_size=%dx%d:pix_fmt=%d:time_base=%d/%d:pixel_aspect=%d/%d",
			pDecodingContext->width, pDecodingContext->height, pDecodingContext->pix_fmt,
			pDecodingContext->time_base.num, pDecodingContext->time_base.den,
			pDecodingContext->sample_aspect_ratio.num,
			pDecodingContext->sample_aspect_ratio.den);

		const AVFilter* pBufferSource = avfilter_get_by_name("buffer");
		if (pBufferSource == nullptr)
		{
			av_log(nullptr, AV_LOG_ERROR, "Video filtering source element not found\n");
			avfilter_graph_free(&pFilterGraph);
			return false;
		}

		int result = avfilter_graph_create_filter(&pBufferSourceContext, pBufferSource, "in", args, nullptr, pFilterGraph);
		if (result < 0)
		{
			av_log(nullptr, AV_LOG_ERROR, "Cannot create video buffer source\n");
			avfilter_graph_free(&pFilterGraph);
			return false;
		}

		const AVFilter* pBufferSink = avfilter_get_by_name("buffersink");
		if (pBufferSink == nullptr)
		{
			av_log(nullptr, AV_LOG_ERROR, "Video filtering sink element not found\n");
			avfilter_graph_free(&pFilterGraph);
			return false;
		}

		result = avfilter_graph_create_filter(&pBufferSinkContext, pBufferSink, "out", nullptr, nullptr, pFilterGraph);
		if (result < 0)
		{
			av_log(nullptr, AV_LOG_ERROR, "Cannot create video buffer sink\n");
			avfilter_graph_free(&pFilterGraph);
			return false;
		}

		result = av_opt_set_bin(pBufferSinkContext, "pix_fmts", (uint8_t*)&pEncodingContext->pix_fmt, 
			sizeof(pEncodingContext->pix_fmt), AV_OPT_SEARCH_CHILDREN);
		if (result < 0)
		{
			av_log(nullptr, AV_LOG_ERROR, "Cannot set output pixel format\n");
			avfilter_graph_free(&pFilterGraph);
			return false;
		}
	}

	if (pDecodingContext->codec_type == AVMEDIA_TYPE_AUDIO)
	{
		if (pDecodingContext->channel_layout == 0)
		{
			pDecodingContext->channel_layout = av_get_default_channel_layout(pDecodingContext->channels);
		}

		char args[512];
		snprintf(args, sizeof(args),
			"time_base=%d/%d:sample_rate=%d:sample_fmt=%s:channel_layout=0x%d",
			pDecodingContext->time_base.num, pDecodingContext->time_base.den, pDecodingContext->sample_rate,
			av_get_sample_fmt_name(pDecodingContext->sample_fmt),
			pDecodingContext->channel_layout);

		const AVFilter* pBufferSource = avfilter_get_by_name("abuffer");
		if (pBufferSource == nullptr)
		{
			av_log(nullptr, AV_LOG_ERROR, "Audio filtering source element not found\n");
			avfilter_graph_free(&pFilterGraph);
			return false;
		}

		int result = avfilter_graph_create_filter(&pBufferSourceContext, pBufferSource, "in", args, NULL, pFilterGraph);
		if (result < 0)
		{
			av_log(nullptr, AV_LOG_ERROR, "Cannot create audio buffer source\n");
			avfilter_graph_free(&pFilterGraph);
			return false;
		}

		const AVFilter* pBufferSink = avfilter_get_by_name("abuffersink");
		if (pBufferSink == nullptr)
		{
			av_log(nullptr, AV_LOG_ERROR, "Audio filtering sink element not found\n");
			avfilter_graph_free(&pFilterGraph);
			return false;
		}

		result = avfilter_graph_create_filter(&pBufferSinkContext, pBufferSink, "out", nullptr, nullptr, pFilterGraph);
		if (result < 0)
		{
			av_log(nullptr, AV_LOG_ERROR, "Cannot create audio buffer sink\n");
			avfilter_graph_free(&pFilterGraph);
			return false;
		}

		result = av_opt_set_bin(pBufferSinkContext, "sample_fmts", 
			(uint8_t*)&pEncodingContext->sample_fmt, sizeof(pEncodingContext->sample_fmt), AV_OPT_SEARCH_CHILDREN);
		if (result < 0)
		{
			av_log(nullptr, AV_LOG_ERROR, "Cannot set output sample format\n");
			avfilter_graph_free(&pFilterGraph);
			return false;
		}

		result = av_opt_set_bin(pBufferSinkContext, "channel_layouts", 
			(uint8_t*)&pEncodingContext->channel_layout, sizeof(pEncodingContext->channel_layout), AV_OPT_SEARCH_CHILDREN);
		if (result < 0)
		{
			av_log(nullptr, AV_LOG_ERROR, "Cannot set output channel layout\n");
			avfilter_graph_free(&pFilterGraph);
			return false;
		}

		result = av_opt_set_bin(pBufferSinkContext, "sample_rates",
			(uint8_t*)&pEncodingContext->sample_rate, sizeof(pEncodingContext->sample_rate), AV_OPT_SEARCH_CHILDREN);
		if (result < 0)
		{
			av_log(nullptr, AV_LOG_ERROR, "Cannot set output sample rate\n");
			avfilter_graph_free(&pFilterGraph);
			return false;
		}
	}

	AVFilterInOut* pOutputs = avfilter_inout_alloc();
	if (pOutputs == nullptr)
	{
		avfilter_graph_free(&pFilterGraph);
		return false;
	}

	pOutputs->name = av_strdup("in");
	if (pOutputs->name == nullptr)
	{
		avfilter_inout_free(&pOutputs);
		avfilter_graph_free(&pFilterGraph);
		return false;
	}

	pOutputs->filter_ctx = pBufferSourceContext;
	pOutputs->pad_idx = 0;
	pOutputs->next = nullptr;

	AVFilterInOut* pInputs = avfilter_inout_alloc();
	if (pInputs == nullptr)
	{
		avfilter_inout_free(&pOutputs);
		avfilter_graph_free(&pFilterGraph);
		return false;
	}

	pInputs->name = av_strdup("out");
	if (pInputs->name == nullptr)
	{
		avfilter_inout_free(&pInputs);
		avfilter_inout_free(&pOutputs);
		avfilter_graph_free(&pFilterGraph);
		return false;
	}

	pInputs->filter_ctx = pBufferSinkContext;
	pInputs->pad_idx = 0;
	pInputs->next = nullptr;

	int result = avfilter_graph_parse_ptr(pFilterGraph, pFilterSpec, &pInputs, &pOutputs, nullptr);
	if (result < 0)
	{
		avfilter_inout_free(&pInputs);
		avfilter_inout_free(&pOutputs);
		avfilter_graph_free(&pFilterGraph);
		return false;
	}

	result = avfilter_graph_config(pFilterGraph, nullptr);
	if (result < 0)
	{
		avfilter_inout_free(&pInputs);
		avfilter_inout_free(&pOutputs);
		avfilter_graph_free(&pFilterGraph);
		return false;
	}

	FilteringContext filteringContext;
	filteringContext.pFilterGraph = pFilterGraph;
	filteringContext.pBufferSourceContext = pBufferSourceContext;
	filteringContext.pBufferSinkContext = pBufferSinkContext;

	m_filteringContexts.push_back(filteringContext);

	avfilter_inout_free(&pInputs);
	avfilter_inout_free(&pOutputs);

	return true;
}

bool LibavStreamer::InitFilters()
{
	if (m_streamContexts.size() != m_pInputFormatContext->nb_streams)
	{
		return false;
	}

	for (int streamIndex = 0; streamIndex < m_pInputFormatContext->nb_streams; streamIndex++)
	{
		if (!(m_pInputFormatContext->streams[streamIndex]->codecpar->codec_type == AVMEDIA_TYPE_AUDIO
			|| m_pInputFormatContext->streams[streamIndex]->codecpar->codec_type == AVMEDIA_TYPE_VIDEO))
		{
			continue;
		}

		const char* pFilterSpec = nullptr;

		if (m_pInputFormatContext->streams[streamIndex]->codecpar->codec_type == AVMEDIA_TYPE_VIDEO)
		{
			//pFilterSpec = "null"; /* passthrough (dummy) filter for video */
			pFilterSpec = "scale=w=iw/2:h=ih/2:sws_flags=fast_bilinear, drawbox=color=0x00000000:t=fill";//"drawbox=color=0x00000000:t=fill";//"scale=w=iw/2:h=ih/2:sws_flags=fast_bilinear";//drawbox=color=0x00000000:t=fill
		}

		if (m_pInputFormatContext->streams[streamIndex]->codecpar->codec_type == AVMEDIA_TYPE_AUDIO)
		{
			//filter_spec = "anull"; /* passthrough (dummy) filter for audio */
			//pFilterSpec = "highpass=f=200, lowpass=f=3000, volume=volume=1.0"; //Primitive noise reduction + volume
			pFilterSpec = "volume=volume=1.0";
		}

		bool result = InitFilter(m_streamContexts[streamIndex].pDecodingContext, m_streamContexts[streamIndex].pEncodingContext, pFilterSpec);
		if (!result)
		{
			return false;
		}

		m_filteringContexts[streamIndex].pEncodedPacket = av_packet_alloc();
		if (m_filteringContexts[streamIndex].pEncodedPacket == nullptr)
		{
			return false;
		}

		m_filteringContexts[streamIndex].pFilteredFrame = av_frame_alloc();
		if (m_filteringContexts[streamIndex].pFilteredFrame == nullptr)
		{
			return false;
		}
	}

	return true;
}

bool LibavStreamer::Volume(double volume)
{
	char strbuf[16];
	snprintf(strbuf, sizeof(strbuf), "%f", volume);

	for (int streamIndex = 0; streamIndex < m_pInputFormatContext->nb_streams; streamIndex++)
	{
		if (m_streamContexts[streamIndex].pEncodingContext == nullptr || 
			m_streamContexts[streamIndex].pDecodingContext == nullptr)
		{
			continue;
		}

		if (m_streamContexts[streamIndex].pDecodingContext->codec_type == AVMEDIA_TYPE_AUDIO && 
			m_streamContexts[streamIndex].pEncodingContext->codec_type == AVMEDIA_TYPE_AUDIO)
		{
			int result = avfilter_graph_send_command(m_filteringContexts[streamIndex].pFilterGraph, "volume", "volume", 
				strbuf, nullptr, 0, 0);
			return result >= 0;
		}
	}

	return false;
}

//bool LibavStreamer::m_isCameraAttached = true;

bool LibavStreamer::Camera(bool attached)
{
	//m_isCameraAttached = attached;

	char colorBlackTransparent[] = "0x00000000";
	char colorBlackOpaque[] = "0x000000FF";

	for (int streamIndex = 0; streamIndex < m_pInputFormatContext->nb_streams; streamIndex++)
	{
		if (m_streamContexts[streamIndex].pEncodingContext == nullptr ||
			m_streamContexts[streamIndex].pDecodingContext == nullptr)
		{
			continue;
		}

		if (m_streamContexts[streamIndex].pDecodingContext->codec_type == AVMEDIA_TYPE_VIDEO &&
			m_streamContexts[streamIndex].pEncodingContext->codec_type == AVMEDIA_TYPE_VIDEO)
		{
			int result = avfilter_graph_send_command(m_filteringContexts[streamIndex].pFilterGraph, "drawbox", "color",
				attached ? colorBlackTransparent : colorBlackOpaque, nullptr, 0, 0);
			return result >= 0;
		}
	}

	return false;
}

bool LibavStreamer::InitFifo()
{
	for (auto&& streamContext : m_streamContexts)
	{
		if (streamContext.pEncodingContext == nullptr || streamContext.pDecodingContext == nullptr)
		{
			continue;
		}

		if (streamContext.pDecodingContext->codec_type == AVMEDIA_TYPE_AUDIO && streamContext.pEncodingContext->codec_type == AVMEDIA_TYPE_AUDIO)
		{
			pFifo = av_audio_fifo_alloc(streamContext.pEncodingContext->sample_fmt, streamContext.pEncodingContext->channels, 1);
			if (pFifo == nullptr)
			{
				av_log(nullptr, AV_LOG_ERROR, "Could not allocate FIFO\n");
				return false;
			}
			return true;
		}
	}
	return false;
}

bool LibavStreamer::FilterEncodeWriteFrame(AVFrame* pFrame, unsigned int stream_index)
{
	FilteringContext filteringContext = m_filteringContexts[stream_index];

	/* push the decoded frame into the filtergraph */
	int result = av_buffersrc_add_frame_flags(filteringContext.pBufferSourceContext, pFrame, 0);
	if (result < 0)
	{
		av_log(nullptr, AV_LOG_ERROR, "Error while feeding the filtergraph\n");
		return false;
	}

	while (true)
	{
		result = av_buffersink_get_frame(filteringContext.pBufferSinkContext, filteringContext.pFilteredFrame);
		if (result < 0)
		{
			/* if no more frames for output - returns AVERROR(EAGAIN)
			 * if flushed and no more frames for output - returns AVERROR_EOF
			 * rewrite retcode to 0 to show it as normal procedure completion
			 */
			if (result == AVERROR(EAGAIN) || result == AVERROR_EOF)
			{
				result = 0;
			}
			break;
		}

		filteringContext.pFilteredFrame->pict_type = AV_PICTURE_TYPE_NONE;

		result = EncodeWriteFrame(stream_index, false);

		av_frame_unref(filteringContext.pFilteredFrame);

		if (!result)
		{
			return false;
		}
	}

	return true;
}

bool LibavStreamer::EncodeWriteFrame(unsigned int stream_index, bool flush)
{
	StreamContext streamContext = m_streamContexts[stream_index];
	FilteringContext filteringContext = m_filteringContexts[stream_index];

	AVFrame* pFilteredFrame = flush ? nullptr : filteringContext.pFilteredFrame;
	
	av_packet_unref(filteringContext.pEncodedPacket);

	int result = avcodec_send_frame(streamContext.pEncodingContext, pFilteredFrame);
	if (result < 0)
	{
		return false;
	}

	while (result >= 0)
	{
		result = avcodec_receive_packet(streamContext.pEncodingContext, filteringContext.pEncodedPacket);
		if (result == AVERROR(EAGAIN) || result == AVERROR_EOF)
		{
			return true;
		}

		filteringContext.pEncodedPacket->stream_index = stream_index;

		av_packet_rescale_ts(filteringContext.pEncodedPacket,
			streamContext.pEncodingContext->time_base,
			m_pOutputFormatContext->streams[stream_index]->time_base);

		result = LibavStreamer::WriteFrame(this, m_pOutputFormatContext, filteringContext.pEncodedPacket);

		av_packet_unref(filteringContext.pEncodedPacket);
	}

	return true;
}

bool LibavStreamer::AddSamplesToFifo(AVAudioFifo* pFifo, uint8_t** convertedInputSamples, const int frameSize)
{
	int error;

	/* Make the FIFO as large as it needs to be to hold both,
	 * the old and the new samples. */
	if ((error = av_audio_fifo_realloc(pFifo, av_audio_fifo_size(pFifo) + frameSize)) < 0) 
	{
		av_log(nullptr, AV_LOG_ERROR, "Could not reallocate FIFO\n");
		return false;
	}

	/* Store the new samples in the FIFO buffer. */
	if (av_audio_fifo_write(pFifo, (void**)convertedInputSamples, frameSize) < frameSize)
	{
		av_log(nullptr, AV_LOG_ERROR, "Could not write data to FIFO\n");
		return false;
	}

	return true;
}

bool LibavStreamer::FilterEncodeWriteFrameAudio(AVFrame* pFrame, unsigned int streamIndex, AVAudioFifo* pFifo)
{
	FilteringContext filteringContext = m_filteringContexts[streamIndex];

	/* push the decoded frame into the filtergraph */
	int result = av_buffersrc_add_frame_flags(filteringContext.pBufferSourceContext, pFrame, 0);
	if (result < 0)
	{
		av_log(nullptr, AV_LOG_ERROR, "Error while feeding the filtergraph\n");
		return false;
	}

	while (true)
	{
		result = av_buffersink_get_frame(filteringContext.pBufferSinkContext, filteringContext.pFilteredFrame);
		if (result < 0)
		{
			/* if no more frames for output - returns AVERROR(EAGAIN)
			 * if flushed and no more frames for output - returns AVERROR_EOF
			 * rewrite retcode to 0 to show it as normal procedure completion
			 */
			if (result == AVERROR(EAGAIN) || result == AVERROR_EOF)
			{
				return true;
			}
		}

		filteringContext.pFilteredFrame->pict_type = AV_PICTURE_TYPE_NONE;

		StreamContext streamContext = m_streamContexts[streamIndex];

		const int outputFrameSize = streamContext.pEncodingContext->frame_size;

		int fifoSize = av_audio_fifo_size(pFifo);

		while (fifoSize < outputFrameSize)
		{
			if (!AddSamplesToFifo(pFifo, filteringContext.pFilteredFrame->extended_data, filteringContext.pFilteredFrame->nb_samples))
			{
				av_frame_unref(filteringContext.pFilteredFrame);
				return false;
			}

			fifoSize = av_audio_fifo_size(pFifo);
		}

		AVRational rational = { 1, streamContext.pEncodingContext->sample_rate };
		int64_t framePts = av_rescale_q(filteringContext.pFilteredFrame->pts, rational, streamContext.pEncodingContext->time_base);

		int numberOfSamples = av_audio_fifo_size(pFifo);
		while (numberOfSamples >= outputFrameSize && numberOfSamples > 0)
		{
			if (!LoadEncodeWrite(pFifo, m_pOutputFormatContext, streamContext.pEncodingContext, streamIndex, framePts))
			{
				av_frame_unref(filteringContext.pFilteredFrame);
				return false;
			}
			numberOfSamples = av_audio_fifo_size(pFifo);
		}

		av_frame_unref(filteringContext.pFilteredFrame);

	}

	return true;
}

namespace
{
	
bool InitOutputFrame(AVFrame** pFrame, AVCodecContext* pOutputCodecContext, int frameSize)
{
	int error;

	if (!(*pFrame = av_frame_alloc()))
	{
		av_log(nullptr, AV_LOG_ERROR, "Could not allocate output frame\n");
		return false;
	}

	(*pFrame)->nb_samples = frameSize;
	(*pFrame)->channel_layout = pOutputCodecContext->channel_layout;
	(*pFrame)->format = pOutputCodecContext->sample_fmt;
	(*pFrame)->sample_rate = pOutputCodecContext->sample_rate;

	if ((error = av_frame_get_buffer(*pFrame, 0)) < 0)
	{
		av_frame_free(pFrame);
		return false;
	}

	return true;
}

}

bool LibavStreamer::LoadEncodeWrite(
	AVAudioFifo* pFifo,
	AVFormatContext* pOutputFormatContext, 
	AVCodecContext* pOutputCodecContext,
	unsigned int streamIndex, 
	int64_t& framePts)
{
	/* Temporary storage of the output samples of the frame written to the file. */
	AVFrame* pOutputFrame = nullptr;

	/* Use the maximum number of possible samples per frame.
	 * If there is less than the maximum possible frame size in the FIFO
	 * buffer use this number. Otherwise, use the maximum possible frame size. */
	const int frameSize = FFMIN(av_audio_fifo_size(pFifo), pOutputCodecContext->frame_size);

	/* Initialize temporary storage for one output frame. */
	if (!InitOutputFrame(&pOutputFrame, pOutputCodecContext, frameSize))
	{
		return false;
	}

	/* Read as many samples from the FIFO buffer as required to fill the frame.
	 * The samples are stored in the frame temporarily. */
	if (av_audio_fifo_read(pFifo, (void**)pOutputFrame->data, frameSize) < frameSize)
	{
		av_log(nullptr, AV_LOG_ERROR, "Could not read data from FIFO\n");
		av_frame_free(&pOutputFrame);
		return false;
	}

	pOutputFrame->pts = framePts;

	/* Encode one frame worth of audio samples. */
	if (!EncodeWriteAudioFrame(streamIndex, pOutputFrame))
	{
		av_frame_free(&pOutputFrame);
		return false;
	}

	framePts += pOutputFrame->nb_samples;

	av_frame_free(&pOutputFrame);

	return true;
}


bool LibavStreamer::EncodeWriteAudioFrame(unsigned int streamIndex, AVFrame* pFrame)
{
	if (pFrame == nullptr)
	{
		return false;
	}

	StreamContext streamContext = m_streamContexts[streamIndex];

	AVPacket* pPacket = av_packet_alloc();
	if (pPacket == nullptr)
	{
		av_log(nullptr, AV_LOG_ERROR, "Could not allocate a packet\n");
		return false;
	}

	FilteringContext filteringContext = m_filteringContexts[streamIndex];

	/* encode filtered frame */
	av_packet_unref(pPacket);

	int result = avcodec_send_frame(streamContext.pEncodingContext, pFrame);
	if (result < 0)
	{
		return false;
	}

	while (result >= 0)
	{
		result = avcodec_receive_packet(streamContext.pEncodingContext, pPacket);
		if (result == AVERROR(EAGAIN) || result == AVERROR_EOF)
		{
			return true;
		}

		pPacket->stream_index = streamIndex;

		av_packet_rescale_ts(pPacket,
			streamContext.pEncodingContext->time_base,
			m_pOutputFormatContext->streams[streamIndex]->time_base);

		result = WriteFrame(this, m_pOutputFormatContext, pPacket);

		av_packet_unref(pPacket);
	}

	av_packet_free(&pPacket);

	return true;
}

bool LibavStreamer::StartDecodingThread()
{
	if (m_packetReceiverThreadStarted)
	{
		return false;
	}

	m_packetReceiverThreadStarted = true;

	m_pPacketReceiverThread = std::make_unique<std::thread>(std::thread(&LibavStreamer::PacketReceiverThreadProc, this));

	return true;
}

bool LibavStreamer::StopDecodingThread()
{
	if (!m_packetReceiverThreadStarted)
	{
		return false;
	}

	m_packetReceiverThreadStarted = false;

	m_pPacketReceiverThread->join();

	return true;
}

bool LibavStreamer::StartVideoEncodingThread()
{
	if (m_videoEncodingThreadStarted)
	{
		return false;
	}

	int streamIndex = 0;
	for (streamIndex = 0; streamIndex < m_streamContexts.size(); ++streamIndex)
	{
		StreamContext streamContext = m_streamContexts[streamIndex];
		if (streamContext.pEncodingContext->codec_type == AVMEDIA_TYPE_VIDEO)
		{
			break;
		}
	}

	if (streamIndex == m_streamContexts.size())
	{
		return false;
	}

	m_videoEncodingThreadStarted = true;

	m_pVideoEncodingThread = std::make_unique<std::thread>(std::thread(&LibavStreamer::VideoDecodingEncodingThreadProc, this, streamIndex));

	return true;
}

bool LibavStreamer::StopVideoEncodingThread()
{
	if (!m_videoEncodingThreadStarted)
	{
		return false;
	}

	m_videoEncodingThreadStarted = false;

	m_pVideoEncodingThread->join();

	return true;
}

bool LibavStreamer::StartAudioEncodingThread()
{
	if (m_audioEncodingThreadStarted)
	{
		return false;
	}

	int streamIndex = 0;
	for (streamIndex = 0; streamIndex < m_streamContexts.size(); ++streamIndex)
	{
		StreamContext streamContext = m_streamContexts[streamIndex];
		if (streamContext.pEncodingContext->codec_type == AVMEDIA_TYPE_AUDIO)
		{
			break;
		}
	}

	if (streamIndex == m_streamContexts.size())
	{
		return false;
	}

	m_audioEncodingThreadStarted = true;

	m_pAudioEncodingThread = std::make_unique<std::thread>(std::thread(&LibavStreamer::AudioDecodingEncodingThreadProc, this, streamIndex));

	return true;
}

bool LibavStreamer::StopAudioEncodingThread()
{
	if (!m_audioEncodingThreadStarted)
	{
		return false;
	}

	m_audioEncodingThreadStarted = false;

	m_pAudioEncodingThread->join();

	return true;
}

void LibavStreamer::FlushFiltersEncoders()
{
	for (int streamIndex = 0; streamIndex < m_pInputFormatContext->nb_streams; streamIndex++)
	{
		FilterEncodeWriteFrame(nullptr, streamIndex);

		if (!(m_streamContexts[streamIndex].pEncodingContext->codec->capabilities & AV_CODEC_CAP_DELAY))
		{
			continue;
		}

		EncodeWriteFrame(streamIndex, true);
	}

	av_write_trailer(m_pOutputFormatContext);
}

int LibavStreamer::WriteFrame(LibavStreamer* pLibavStreamer, AVFormatContext* pOutputFormatContext, AVPacket* pPacket)
{
	std::lock_guard<std::mutex> guard(pLibavStreamer->m_writeFrameMutex);
	return av_write_frame(pOutputFormatContext, pPacket);
	//return av_interleaved_write_frame(pOutputFormatContext, pPacket);
	//return 0;
}

void LibavStreamer::PacketReceiverThreadProc(LibavStreamer* pLibavStreamer)
{
	::CoInitializeEx(nullptr, COINIT_MULTITHREADED);

	while (pLibavStreamer->m_packetReceiverThreadStarted)
	{
		AVPacket* pPacket = av_packet_alloc();
		if (pPacket == nullptr)
		{
			av_log(nullptr, AV_LOG_ERROR, "Error allocating a packet.\n");
			break;
		}

		int result = av_read_frame(pLibavStreamer->m_pInputFormatContext, pPacket);
		if (result < 0)
		{
			av_log(nullptr, AV_LOG_ERROR, "Error reading frame\n");
			av_packet_free(&pPacket);
			break;
		}

		unsigned int streamIndex = pPacket->stream_index;

		if (pLibavStreamer->m_filteringContexts.size() <= streamIndex || pLibavStreamer->m_streamContexts.size() <= streamIndex)
		{
			av_log(nullptr, AV_LOG_ERROR, "No context or filter for stream index\n");
			av_packet_free(&pPacket);
			break;
		}

		FilteringContext filteringContext = pLibavStreamer->m_filteringContexts[streamIndex];
		if (filteringContext.pFilterGraph == nullptr)
		{
			av_log(nullptr, AV_LOG_ERROR, "No filter graph for stream index\n");
			av_packet_free(&pPacket);
			break;
		}

		StreamContext streamContext = pLibavStreamer->m_streamContexts[streamIndex];

		av_packet_rescale_ts(pPacket,
			pLibavStreamer->m_pInputFormatContext->streams[streamIndex]->time_base,
			streamContext.pDecodingContext->time_base);

		streamContext.pPacketsQueue->enqueue(pPacket);
	}

	::CoUninitialize();
}

void LibavStreamer::VideoDecodingEncodingThreadProc(LibavStreamer* pLibavStreamer, int streamIndex)
{
	::CoInitializeEx(nullptr, COINIT_MULTITHREADED);

	StreamContext videoStreamContext = pLibavStreamer->m_streamContexts[streamIndex];
	while (pLibavStreamer->m_videoEncodingThreadStarted)
	{
		AVPacket* pPacket = nullptr;
		bool dequeueResult = videoStreamContext.pPacketsQueue->try_dequeue<AVPacket*>(pPacket);
		if (!dequeueResult)
		{
			std::this_thread::sleep_for(std::chrono::duration<double, std::milli>(30));
			continue;
		}

		/*if (!m_isCameraAttached)
		{
			av_packet_free(&pPacket);
			continue;
		}*/

		int result = avcodec_send_packet(videoStreamContext.pDecodingContext, pPacket);
		if (result < 0)
		{
			av_log(nullptr, AV_LOG_ERROR, "Decoding failed\n");
			av_packet_free(&pPacket);
			break;
		}

		while (result >= 0)
		{
			AVFrame* pDecodedFrame = av_frame_alloc();
			if (pDecodedFrame == nullptr)
			{
				av_log(nullptr, AV_LOG_ERROR, "Failed to allocate a frame for stream #%u\n", streamIndex);
				break;
			}

			result = avcodec_receive_frame(videoStreamContext.pDecodingContext, pDecodedFrame);
			if (result == AVERROR_EOF || result == AVERROR(EAGAIN))
			{
				av_frame_free(&pDecodedFrame);
				break;
			}
			if (result < 0)
			{
				av_log(nullptr, AV_LOG_ERROR, "Error receiving frame from decoder\n");
				av_frame_free(&pDecodedFrame);
				break;
			}

			pDecodedFrame->pts = pDecodedFrame->best_effort_timestamp;

			pLibavStreamer->FilterEncodeWriteFrame(pDecodedFrame, streamIndex);

			av_frame_free(&pDecodedFrame);
		}

		av_packet_free(&pPacket);
	}

	::CoUninitialize();
}

void LibavStreamer::AudioDecodingEncodingThreadProc(LibavStreamer* pLibavStreamer, int streamIndex)
{
	::CoInitializeEx(nullptr, COINIT_MULTITHREADED);

	StreamContext audioStreamContext = pLibavStreamer->m_streamContexts[streamIndex];
	while (pLibavStreamer->m_audioEncodingThreadStarted)
	{
		AVPacket* pPacket = nullptr;
		bool dequeueResult = audioStreamContext.pPacketsQueue->try_dequeue<AVPacket*>(pPacket);
		if (!dequeueResult)
		{
			std::this_thread::sleep_for(std::chrono::duration<double, std::milli>(40));
			continue;
		}

		int result = avcodec_send_packet(audioStreamContext.pDecodingContext, pPacket);
		if (result < 0)
		{
			av_log(nullptr, AV_LOG_ERROR, "Decoding failed\n");
			av_packet_free(&pPacket);
			break;
		}

		while (result >= 0)
		{
			AVFrame* pDecodedFrame = av_frame_alloc();
			if (pDecodedFrame == nullptr)
			{
				av_log(nullptr, AV_LOG_ERROR, "Failed to allocate a frame for stream #%u\n", streamIndex);
				break;
			}

			result = avcodec_receive_frame(audioStreamContext.pDecodingContext, pDecodedFrame);
			if (result == AVERROR_EOF || result == AVERROR(EAGAIN))
			{
				av_frame_free(&pDecodedFrame);
				break;
			}
			if (result < 0)
			{
				av_log(nullptr, AV_LOG_ERROR, "Error receiving frame from decoder\n");
				av_frame_free(&pDecodedFrame);
				break;
			}

			pDecodedFrame->pts = pDecodedFrame->best_effort_timestamp;

			pLibavStreamer->FilterEncodeWriteFrameAudio(pDecodedFrame, streamIndex, pLibavStreamer->pFifo);

			av_frame_free(&pDecodedFrame);
		}

		av_packet_free(&pPacket);
	}

	::CoUninitialize();
}
